#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
from __future__ import annotations

import unittest
from dataclasses import dataclass
from typing import Union, List, Any

# @manual=//pytorch/torcharrow/csrc/velox:_torcharrow
import torcharrow._torcharrow as ta


class BaseTestColumns(unittest.TestCase):
    def assert_Column(self, col: ta.BaseColumn, val: List[Any]):
        self.assertEqual(len(col), len(val))
        self.assertEqual(col.get_null_count(), sum(x is None for x in val))
        for i in range(len(val)):
            if val[i] is None:
                self.assertTrue(col.is_null_at(i))
            else:
                self.assertFalse(col.is_null_at(i))
                if isinstance(val[i], list):
                    self.assert_Column(col[i], val[i])
                elif isinstance(val[i], float):
                    self.assertAlmostEqual(col[i], val[i], places=6)
                else:
                    self.assertEqual(col[i], val[i])


class TestSimpleColumns(BaseTestColumns):
    def test_SimpleColumnInt64(self):
        data = [1, 2, None, 3, 4, None]
        col = infer_column(data)

        self.assertEqual(col[0], 1)
        self.assertEqual(col[1], 2)
        self.assertEqual(col[3], 3)
        self.assertEqual(col[4], 4)

        self.assertEqual(len(col), 6)

        with self.assertRaises(TypeError):
            # TypeError: an integer is required (got type NoneType)
            col.append(None)

        with self.assertRaises(TypeError):
            # TypeError: an integer is required (got type String)
            col.append("hello")

        self.assertEqual(col.is_null_at(0), False)
        self.assertEqual(col.is_null_at(1), False)
        self.assertEqual(col.is_null_at(2), True)
        self.assertEqual(col.is_null_at(3), False)
        self.assertEqual(col.is_null_at(4), False)
        self.assertEqual(col.is_null_at(5), True)
        self.assertEqual(col.get_null_count(), 2)

        sliced_col = col.slice(1, 3)
        self.assertEqual(len(sliced_col), 3)
        self.assertEqual(sliced_col[0], 2)
        self.assertEqual(sliced_col[2], 3)
        self.assertEqual(sliced_col.get_null_count(), 1)

    def test_SimpleColumnInt64_unary(self):
        data = [1, -2, None, 3, -4, None]
        col = infer_column(data)
        self.assertEqual(col.type().kind(), ta.TypeKind.BIGINT)

        neg_col = col.neg()
        self.assert_Column(neg_col, [-1, 2, None, -3, 4, None])
        self.assertEqual(neg_col.type().kind(), ta.TypeKind.BIGINT)

        neg_col2 = neg_col.neg()
        self.assert_Column(neg_col2, [1, -2, None, 3, -4, None])
        self.assertEqual(neg_col2.type().kind(), ta.TypeKind.BIGINT)

        neg_col3 = neg_col2.neg()
        self.assert_Column(neg_col3, [-1, 2, None, -3, 4, None])
        self.assertEqual(neg_col3.type().kind(), ta.TypeKind.BIGINT)

        abs_col = col.abs()
        self.assert_Column(abs_col, [1, 2, None, 3, 4, None])
        self.assertEqual(abs_col.type().kind(), ta.TypeKind.BIGINT)

    def test_SimpleColumnInt64_binary(self):
        data1 = [1, -2, None, 3, -4, None]
        col1 = infer_column(data1)
        data2 = [None, 1, 2, 3, 4, 5]
        col2 = infer_column(data2)

        # column ops
        sum_col = col1.add(col2)
        self.assert_Column(sum_col, [None, -1, None, 6, 0, None])
        self.assertEqual(sum_col.type().kind(), ta.TypeKind.BIGINT)

        self.assert_Column(col1.sub(col2), [None, -3, None, 0, -8, None])
        self.assert_Column(col1.mul(col2), [None, -2, None, 9, -16, None])
        self.assert_Column(col1.mod(col2), [None, 0, None, 0, 0, None])

        # type promotion
        data3 = [None, 1.0, 2.0, 3.0, 4.0, 5.0]
        col3 = infer_column(data3)
        self.assertEqual(col3.type().kind(), ta.TypeKind.REAL)

        sum_col = col1.add(col3)
        self.assertEqual(sum_col.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(sum_col, [None, -1.0, None, 6.0, 0.0, None])

        sum_col2 = col3.add(col1)
        self.assertEqual(sum_col2.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(sum_col2, [None, -1.0, None, 6.0, 0.0, None])

        # scalar ops
        add_scalar = col1.add(1)
        self.assertEqual(add_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(add_scalar, [2, -1, None, 4, -3, None])

        add_scalar = col1.add(0.1)
        self.assertEqual(add_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(add_scalar, [1.1, -1.9, None, 3.1, -3.9, None])

        add_scalar = col1.radd(1)
        self.assertEqual(add_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(add_scalar, [2, -1, None, 4, -3, None])

        add_scalar = col1.radd(0.1)
        self.assertEqual(add_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(add_scalar, [1.1, -1.9, None, 3.1, -3.9, None])

        sub_scalar = col1.sub(2)
        self.assertEqual(sub_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(sub_scalar, [-1, -4, None, 1, -6, None])

        sub_scalar = col1.sub(0.1)
        self.assertEqual(sub_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(sub_scalar, [0.9, -2.1, None, 2.9, -4.1, None])

        sub_scalar = col1.rsub(2)
        self.assertEqual(sub_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(sub_scalar, [1, 4, None, -1, 6, None])

        sub_scalar = col1.rsub(0.1)
        self.assertEqual(sub_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(sub_scalar, [-0.9, 2.1, None, -2.9, 4.1, None])

        mul_scalar = col1.mul(2)
        self.assertEqual(mul_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(mul_scalar, [2, -4, None, 6, -8, None])

        mul_scalar = col1.mul(-2.0)
        self.assertEqual(mul_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(mul_scalar, [-2.0, 4.0, None, -6.0, 8.0, None])

        mul_scalar = col1.rmul(2)
        self.assertEqual(mul_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(mul_scalar, [2, -4, None, 6, -8, None])

        mul_scalar = col1.rmul(-2.0)
        self.assertEqual(mul_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(mul_scalar, [-2.0, 4.0, None, -6.0, 8.0, None])

        mod_scalar = col1.mod(3)
        self.assertEqual(mod_scalar.type().kind(), ta.TypeKind.BIGINT)
        # Python's behavior for modulo on negative numbers is different from
        # Velox/C++. e.g. -2 % 3 = 1 (because integer division in Python is
        # floor(a/b), as opposed to trunc(a/b))
        self.assert_Column(mod_scalar, [1, -2, None, 0, -1, None])

        mod_scalar = col1.mod(-3.0)
        self.assertEqual(mod_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(mod_scalar, [1.0, -2.0, None, 0.0, -1.0, None])

        mod_scalar = col1.rmod(3)
        self.assertEqual(mod_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(mod_scalar, [0, 1, None, 0, 3, None])

        mod_scalar = col1.rmod(-3.0)
        self.assertEqual(mod_scalar.type().kind(), ta.TypeKind.REAL)
        self.assert_Column(mod_scalar, [0.0, -1.0, None, 0.0, -3.0, None])

        # It's debatable whether this (add BIGINT with BOOLEAN) should be supported.
        # But since PyTorch supports it for NumPy compatbility, TorchArrow also supports this.
        add_scalar = col1.add(True)
        self.assertEqual(add_scalar.type().kind(), ta.TypeKind.BIGINT)
        self.assert_Column(add_scalar, [2, -1, None, 4, -3, None])

    def test_SimpleColumnFloat32_unary(self):
        data = [1.2, -2.3, None, 3.4, -4.6, None]
        col = infer_column(data)
        self.assertEqual(col.type().kind(), ta.TypeKind.REAL)

        neg_col = col.neg()
        self.assert_Column(neg_col, [-1.2, 2.3, None, -3.4, 4.6, None])
        self.assertEqual(neg_col.type().kind(), ta.TypeKind.REAL)

        abs_col = col.abs()
        self.assert_Column(abs_col, [1.2, 2.3, None, 3.4, 4.6, None])
        self.assertEqual(abs_col.type().kind(), ta.TypeKind.REAL)

        round_col = col.round()
        self.assert_Column(round_col, [1.0, -2.0, None, 3.0, -5.0, None])
        self.assertEqual(round_col.type().kind(), ta.TypeKind.REAL)

    def test_SimpleColumnBoolean(self):
        data = [True, True, True, True]
        col = infer_column(data)

        for i in range(4):
            self.assertEqual(col[i], True)

        self.assertEqual(len(col), 4)

        with self.assertRaises(TypeError):
            # TypeError: a boolean is required (got type NoneType)
            col.append(None)

        with self.assertRaises(TypeError):
            # TypeError: a boolean is required (got type String)
            col.append("hello")

        col.append_null()
        self.assertEqual(col.is_null_at(0), False)
        self.assertEqual(col.is_null_at(1), False)
        self.assertEqual(col.is_null_at(2), False)
        self.assertEqual(col.is_null_at(3), False)
        self.assertEqual(col.is_null_at(4), True)

    def test_SimpleColumnBoolean_unary(self):
        data = [True, False, None, True, False, None]
        col = infer_column(data)
        self.assertEqual(col.type().kind(), ta.TypeKind.BOOLEAN)

        inv_col = col.invert()
        self.assertEqual(inv_col.type().kind(), ta.TypeKind.BOOLEAN)
        self.assert_Column(inv_col, [False, True, None, False, True, None])

    def test_SimpleColumnString(self):
        data = ["0", "1", "2", "3"]
        col = infer_column(data)

        for i in range(4):
            self.assertEqual(col[i], str(i))

        self.assertEqual(len(col), 4)

        with self.assertRaises(TypeError):
            # TypeError: a string is required (got type NoneType)
            col.append(None)

        with self.assertRaises(TypeError):
            # TypeError: a string is required (got type int)
            col.append(1)

        col.append_null()
        self.assertEqual(col.is_null_at(0), False)
        self.assertEqual(col.is_null_at(1), False)
        self.assertEqual(col.is_null_at(2), False)
        self.assertEqual(col.is_null_at(3), False)
        self.assertEqual(col.is_null_at(4), True)

    def test_SimpleColumnString_unary(self):
        data = ["abc", "ABC", "XYZ123", None, "xYZ", "123", "äöå", ",.!"]
        col = infer_column(data)

        lcol = col.lower()
        self.assert_Column(
            lcol, ["abc", "abc", "xyz123", None, "xyz", "123", "äöå", ",.!"]
        )

        ucol = col.upper()
        self.assert_Column(
            ucol, ["ABC", "ABC", "XYZ123", None, "XYZ", "123", "ÄÖÅ", ",.!"]
        )

        lcol2 = ucol.lower()
        self.assert_Column(
            lcol2, ["abc", "abc", "xyz123", None, "xyz", "123", "äöå", ",.!"]
        )

        ucol2 = lcol.upper()
        self.assert_Column(
            ucol2, ["ABC", "ABC", "XYZ123", None, "XYZ", "123", "ÄÖÅ", ",.!"]
        )

        alpha = col.isalpha()
        self.assert_Column(alpha, [True, True, False, None, True, False, True, False])

        alnum = col.isalnum()
        self.assert_Column(alnum, [True, True, True, None, True, True, True, False])

        integer = col.isinteger()
        self.assert_Column(
            integer, [False, False, False, None, False, True, False, False]
        )

    def test_SimpleColumnString_islower(self):
        data = {
            "abc": True,
            "": False,
            " ": False,
            "a  ": True,
            "123": False,
            "a123": True,
            ",.!": False,
            "a,.!": True,
            "ABC": False,
            "abcXyz": False,
            None: None,
            "äöå": True,
            "äöå  123 ,.!": True,
            "ÄÖÅ": False,
            "öÄå": False,
            "♬ ♫ ♯": False,
            "abc ♬ ♫ ♯": True,
        }
        col = infer_column(list(data.keys()))
        islower = col.islower()
        self.assert_Column(islower, list(data.values()))

    def test_SimpleColumnUTF(self):
        s = ["hello.this", "is.interesting.", "this.is_24", "paradise"]
        col = infer_column(s)
        for i in range(4):
            self.assertEqual(col[i], s[i])

        self.assertEqual(len(col), 4)

    def test_ConstantColumn(self):
        ###########
        #  BIGINT
        col = ta.ConstantColumn(42, 6)
        self.assertTrue(isinstance(col.type(), ta.VeloxType_BIGINT))
        self.assert_Column(col, [42] * 6)

        # Test use constant column for normal add
        data = [1, -2, None, 3, -4, None]
        num_column = infer_column(data)
        add_result = num_column.add(col)
        self.assertTrue(isinstance(add_result.type(), ta.VeloxType_BIGINT))
        self.assert_Column(add_result, [43, 40, None, 45, 38, None])

        add_result = col.add(num_column)
        self.assertTrue(isinstance(add_result.type(), ta.VeloxType_BIGINT))
        self.assert_Column(add_result, [43, 40, None, 45, 38, None])

        ###########
        #  REAL
        col = ta.ConstantColumn(4.2, 6)
        self.assertTrue(isinstance(col.type(), ta.VeloxType_REAL))
        self.assert_Column(col, [4.2] * 6)

        # Test use constant column for normal add
        data = [1.2, -2.3, None, 3.4, -4.6, None]
        num_column = infer_column(data)
        add_result = num_column.add(col)
        self.assertTrue(isinstance(add_result.type(), ta.VeloxType_REAL))
        self.assert_Column(add_result, [5.4, 1.9, None, 7.6, -0.4, None])

        add_result = col.add(num_column)
        self.assertTrue(isinstance(add_result.type(), ta.VeloxType_REAL))
        self.assert_Column(add_result, [5.4, 1.9, None, 7.6, -0.4, None])

        ###########
        #  VARCHAR
        col = ta.ConstantColumn("abc", 6)
        self.assertTrue(isinstance(col.type(), ta.VeloxType_VARCHAR))
        self.assert_Column(col, ["abc"] * 6)

    def test_FromPyList(self):
        #  BIGINT
        col = ta.Column(ta.VeloxType_BIGINT(), [1, 2, None, 4])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_BIGINT))
        self.assert_Column(col, [1, 2, None, 4])

        #  INTEGER
        col = ta.Column(ta.VeloxType_INTEGER(), [1, 2, None, 4])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_INTEGER))
        self.assert_Column(col, [1, 2, None, 4])

        #  SMALLINT
        col = ta.Column(ta.VeloxType_SMALLINT(), [1, 2, None, 4])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_SMALLINT))
        self.assert_Column(col, [1, 2, None, 4])

        #  TINYINT
        col = ta.Column(ta.VeloxType_TINYINT(), [1, 2, None, 4])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_TINYINT))
        self.assert_Column(col, [1, 2, None, 4])

        #  REAL
        col = ta.Column(ta.VeloxType_REAL(), [1, 2, None, 4])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_REAL))
        self.assert_Column(col, [1.0, 2.0, None, 4.0])

        #  DOUBLE
        col = ta.Column(ta.VeloxType_DOUBLE(), [1, 2, None, 4])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_DOUBLE))
        self.assert_Column(col, [1.0, 2.0, None, 4.0])

        #  BOOLEAN
        col = ta.Column(ta.VeloxType_BOOLEAN(), [True, False, None, True])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_BOOLEAN))
        self.assert_Column(col, [True, False, None, True])

        #  VARCHAR
        col = ta.Column(ta.VeloxType_VARCHAR(), ["foo", "bar", None, "abc"])
        self.assertTrue(isinstance(col.type(), ta.VeloxType_VARCHAR))
        self.assert_Column(col, ["foo", "bar", None, "abc"])

        #  ARRAY of scalar element
        col = ta.Column(
            ta.VeloxArrayType(ta.VeloxType_VARCHAR()),
            [["foo", "bar"], None, ["abc", None]],
        )
        self.assertTrue(isinstance(col.type(), ta.VeloxArrayType))
        self.assertTrue(isinstance(col.type().element_type(), ta.VeloxType_VARCHAR))
        self.assert_Column(col, [["foo", "bar"], None, ["abc", None]])

        #  ARRAY of ROW element
        # col = ta.Column(
        #     ta.VeloxArrayType(
        #         ta.VeloxRowType(
        #             ["f1", "f2"], [ta.VeloxType_VARCHAR(), ta.VeloxType_BIGINT()]
        #         )
        #     ),
        #     [[("foo", 1), ("bar", 2)], None, [("abc", 3), ("def", 4)]],
        # )
        # self.assertTrue(isinstance(col.type(), ta.VeloxArrayType))
        # self.assertTrue(isinstance(col.type().element_type(), ta.VeloxRowType))
        # self.assert_Column(
        #     col, [[("foo", 1), ("bar", 2)], None, [("abc", 3), ("def", 4)]]
        # )

    def test_NullCount(self):
        col = infer_column([None, 1, 2, None])
        self.assertEqual(col.get_null_count(), 2)

        colSlice = col.slice(col.offset, col.length)
        self.assertEqual(colSlice.get_null_count(), 2)

        colSlice = col.slice(1, col.length - 1)
        self.assertEqual(colSlice.get_null_count(), 1)
        self.assertEqual(col.get_null_count(), 2)

        colSlice = col.slice(1, 2)
        self.assertEqual(colSlice.get_null_count(), 0)
        self.assertEqual(col.get_null_count(), 2)


def is_same_type(a, b) -> bool:
    if isinstance(a, ta.VeloxType_BIGINT):
        return isinstance(b, ta.VeloxType_BIGINT)
    if isinstance(a, ta.VeloxType_VARCHAR):
        return isinstance(b, ta.VeloxType_VARCHAR)
    if isinstance(a, ta.VeloxType_BOOLEAN):
        return isinstance(b, ta.VeloxType_BOOLEAN)
    if isinstance(a, ta.VeloxArrayType):
        return isinstance(b, ta.VeloxArrayType) and is_same_type(
            a.element_type(), b.element_type()
        )
    raise NotImplementedError()


# infer result
@dataclass(frozen=True)
class Unresolved:
    def union(self, other: Unresolved) -> Unresolved:
        return other


@dataclass(frozen=True)
class UnresolvedArray(Unresolved):
    element_type: Unresolved


def infer_column(data) -> ta.BaseColumn:
    inferred_column = _infer_column(data)
    if isinstance(inferred_column, Unresolved):
        return resolve_column_with_arbitrary_type(inferred_column)
    else:
        return inferred_column


def resolve_column_with_arbitrary_type(unresolved: Unresolved) -> ta.BaseColumn:
    if isinstance(unresolved, UnresolvedArray):
        element = resolve_column_with_arbitrary_type(unresolved.element_type)
        col = ta.Column(ta.VeloxArrayType(element.type()))
        col.append(element)
        return col
    else:
        return ta.Column(ta.VeloxType_BIGINT())


def get_union_type(inferred_columns: List[Union[ta.BaseColumn, Unresolved, None]]):
    unresolved_item_type = None
    resolved_item_type = None
    for item_col in inferred_columns:
        if item_col is None:
            pass
        elif isinstance(item_col, Unresolved):
            if unresolved_item_type is None:
                unresolved_item_type = item_col
            else:
                unresolved_item_type = unresolved_item_type.union(item_col)

        elif resolved_item_type is None:
            resolved_item_type = item_col.type()
        else:
            assert is_same_type(resolved_item_type, item_col.type())

    if resolved_item_type is None:
        if unresolved_item_type is None:
            return None
        else:
            return unresolved_item_type
    else:
        return resolved_item_type


def _infer_column(data) -> Union[ta.BaseColumn, Unresolved, None]:
    if data is None:
        return None
    assert isinstance(data, list)
    non_null_item = next((item for item in data if item is not None), None)

    if non_null_item is None:
        return Unresolved()
    else:
        if isinstance(non_null_item, list):
            inferred_columns = [_infer_column(item) for item in data]
            union_type = get_union_type(inferred_columns)
            if union_type is None:
                return Unresolved()
            elif isinstance(union_type, Unresolved):
                return UnresolvedArray(union_type)
            else:
                resolved_item_type = union_type
                col = ta.Column(ta.VeloxArrayType(resolved_item_type))
                for item_col, item in zip(inferred_columns, data):
                    if item is None:
                        resolved_item_col = None
                    elif isinstance(item_col, Unresolved):
                        resolved_item_col = resolve_column(item, resolved_item_type)
                    else:
                        resolved_item_col = item_col

                    if resolved_item_col is None:
                        col.append_null()
                    else:
                        col.append(resolved_item_col)
                return col
        elif isinstance(non_null_item, dict):
            keys_array = []
            values_array = []
            for item in data:
                if item is None:
                    keys_array.append(None)
                    values_array.append(None)
                elif isinstance(item, dict):
                    keys_array.append(list(item.keys()))
                    values_array.append(list(item.values()))
                else:
                    raise ValueError("non-dict item in dict list")

            inferred_keys_array_columns = _infer_column(keys_array)
            inferred_values_array_columns = _infer_column(values_array)

            keys_array_type = inferred_keys_array_columns.type()
            values_array_type = inferred_values_array_columns.type()

            if isinstance(keys_array_type, ta.VeloxArrayType) and isinstance(
                values_array_type, ta.VeloxArrayType
            ):
                col = ta.Column(
                    ta.VeloxMapType(
                        keys_array_type.element_type(), values_array_type.element_type()
                    )
                )
                for item in data:
                    if item is None:
                        col.append_null()
                    else:
                        key_col = ta.Column(keys_array_type.element_type())
                        value_col = ta.Column(values_array_type.element_type())
                        for key, value in item.items():
                            key_col.append(key)
                            if value is None:
                                value_col.append_null()
                            else:
                                value_col.append(value)
                        col.append(key_col, value_col)
                return col
            else:
                raise NotImplementedError()

        else:
            type_ = {
                int: ta.VeloxType_BIGINT(),
                float: ta.VeloxType_REAL(),
                str: ta.VeloxType_VARCHAR(),
                bool: ta.VeloxType_BOOLEAN(),
            }.get(type(non_null_item))
            if type_ is None:
                raise NotImplementedError(f"Cannot infer {type(non_null_item)}")
            else:
                col = ta.Column(type_)
                for item in data:
                    if item is None:
                        col.append_null()
                    else:
                        col.append(item)
                return col


def resolve_column(item, type_) -> ta.BaseColumn:
    col = ta.Column(type_)
    for value in item:
        if value is None:
            col.append_null()
        else:
            if type(type_) in (
                ta.VeloxType_INTEGER,
                ta.VeloxType_VARCHAR,
                ta.VeloxType_BOOLEAN,
            ):
                col.append(value)
            elif type(type_) == ta.VeloxArrayType:
                col.append(resolve_column(value, type_.element_type()))
            else:
                raise NotImplementedError(f"{type(type_)}")
    return col


class TestInferColumn(unittest.TestCase):
    def test_infer_simple(self):
        data = [1, 2, 3]
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxType_BIGINT()))

    def test_infer_array(self):
        data = [[1], [2], [3]]
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxArrayType(ta.VeloxType_BIGINT())))

    def test_infer_nested_array(self):
        data = [[[1]], [[2], [5]], [[3, 4]]]
        type_ = infer_column(data).type()
        self.assertTrue(
            is_same_type(
                type_, ta.VeloxArrayType(ta.VeloxArrayType(ta.VeloxType_BIGINT()))
            )
        )

    def test_unresolved(self):
        data = []
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxType_BIGINT()))

    def test_nested_unresolved1(self):
        data = [[]]
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxArrayType(ta.VeloxType_BIGINT())))

    def test_nested_unresolved2(self):
        data = [None]
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxType_BIGINT()))

    def test_nested_unresolved3(self):
        data = [[None]]
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxArrayType(ta.VeloxType_BIGINT())))

    def test_propagate_unresolved(self):
        data = [None, [], [1], [1, None, 2], None]
        type_ = infer_column(data).type()
        self.assertTrue(is_same_type(type_, ta.VeloxArrayType(ta.VeloxType_BIGINT())))


class TestArrayColumns(BaseTestColumns):
    def test_ArrayColumnInt64(self):
        data = [None, [], [1], [1, None, 2], None]
        col = infer_column(data)

        self.assert_Column(col.elements(), [1, 1, None, 2])

        for sliced_col, sliced_data in (
            (col, data),
            (col.slice(2, 2), data[2:4]),
            (col.slice(1, 4), data[1:5]),
        ):
            self.assertEqual(len(sliced_col), len(sliced_data))
            for i, item in enumerate(sliced_data):
                if item is None:
                    self.assertTrue(sliced_col.is_null_at(i))
                else:
                    self.assertFalse(sliced_col.is_null_at(i))
                    self.assertEqual(len(sliced_col[i]), len(item))
                    for j, value in enumerate(item):
                        if value is None:
                            self.assertTrue(sliced_col[i].is_null_at(j))
                        else:
                            self.assertFalse(sliced_col[i].is_null_at(j))
                            self.assertEqual(sliced_col[i][j], sliced_data[i][j])

    def test_NestedArrayColumnInt64(self):
        data = [[[1, 2], None, [3, 4]], [[4], [5]]]
        col = infer_column(data)
        self.assertEqual(col[0][0][0], 1)
        self.assertEqual(col[0][0][1], 2)
        self.assertTrue(col[0].is_null_at(1))
        self.assertEqual(col[0][2][0], 3)
        self.assertEqual(col[0][2][1], 4)
        self.assertEqual(col[1][0][0], 4)
        self.assertEqual(col[1][1][0], 5)

    def test_NestedArrayColumnString(self):
        data = [[], [[]], [["a"]], [["b", "c"], ["d", "e", "f"]]]
        col = infer_column(data)
        self.assertEqual(len(col[0]), 0)
        self.assertEqual(len(col[1]), 1)
        self.assertEqual(len(col[1][0]), 0)
        self.assertEqual(col[2][0][0], "a")
        self.assertEqual(col[3][0][0], "b")
        self.assertEqual(col[3][0][1], "c")
        self.assertEqual(col[3][1][0], "d")
        self.assertEqual(col[3][1][1], "e")
        self.assertEqual(col[3][1][2], "f")


class TestMapColumns(unittest.TestCase):
    def test_MapColumnInt64(self):
        data = [{"a": 1, "b": 2}, {"c": 3, "d": 4, "e": 5}]
        col = infer_column(data)
        self.assertEqual(len(col), 2)
        keys = col.keys()
        self.assertEqual(len(keys), 2)
        self.assertEqual(len(keys[0]), 2)
        self.assertEqual(keys[0][0], "a")
        self.assertEqual(keys[0][1], "b")
        self.assertEqual(len(keys[1]), 3)
        self.assertEqual(keys[1][0], "c")
        self.assertEqual(keys[1][1], "d")
        self.assertEqual(keys[1][2], "e")
        values = col.values()
        self.assertEqual(len(values), 2)
        self.assertEqual(len(values[0]), 2)
        self.assertEqual(values[0][0], 1)
        self.assertEqual(values[0][1], 2)
        self.assertEqual(len(values[1]), 3)
        self.assertEqual(values[1][0], 3)
        self.assertEqual(values[1][1], 4)
        self.assertEqual(values[1][2], 5)

        sliced_col = col.slice(1, 1)
        self.assertEqual(len(sliced_col), 1)
        keys = sliced_col.keys()
        self.assertEqual(len(keys), 1)
        self.assertEqual(len(keys[0]), 3)
        self.assertEqual(keys[0][0], "c")
        self.assertEqual(keys[0][1], "d")
        self.assertEqual(keys[0][2], "e")
        values = sliced_col.values()
        self.assertEqual(len(values), 1)
        self.assertEqual(len(values[0]), 3)
        self.assertEqual(values[0][0], 3)
        self.assertEqual(values[0][1], 4)
        self.assertEqual(values[0][2], 5)

    def test_MapColumnInt64_with_none(self):
        data = [None, {"a": 1, "b": 2}, {"c": None, "d": 4, "e": 5}]
        col = infer_column(data)
        self.assertEqual(len(col), 3)
        self.assertTrue(col.is_null_at(0))
        keys = col.keys()
        self.assertEqual(len(keys), 3)
        self.assertEqual(len(keys[1]), 2)
        self.assertEqual(keys[1][0], "a")
        self.assertEqual(keys[1][1], "b")
        self.assertEqual(len(keys[2]), 3)
        self.assertEqual(keys[2][0], "c")
        self.assertEqual(keys[2][1], "d")
        self.assertEqual(keys[2][2], "e")
        values = col.values()
        self.assertEqual(len(values), 3)
        self.assertEqual(len(values[1]), 2)
        self.assertEqual(values[1][0], 1)
        self.assertEqual(values[1][1], 2)
        self.assertEqual(len(values[2]), 3)
        self.assertTrue(values[2].is_null_at(0))
        self.assertEqual(values[2][1], 4)
        self.assertEqual(values[2][2], 5)

        sliced_col = col.slice(1, 1)
        self.assertEqual(len(sliced_col), 1)
        keys = sliced_col.keys()
        self.assertEqual(len(keys), 1)
        self.assertEqual(len(keys[0]), 2)
        self.assertEqual(keys[0][0], "a")
        self.assertEqual(keys[0][1], "b")
        values = sliced_col.values()
        self.assertEqual(len(values), 1)
        self.assertEqual(len(values[0]), 2)
        self.assertEqual(values[0][0], 1)
        self.assertEqual(values[0][1], 2)


class TestRowColumns(unittest.TestCase):
    def test_RowColumn1(self):
        col = ta.Column(
            ta.VeloxRowType(
                ["a", "b"], [ta.VeloxType_INTEGER(), ta.VeloxType_VARCHAR()]
            )
        )
        col.child_at(0).append(1)
        col.child_at(1).append("x")
        col.set_length(1)
        col.child_at(0).append(2)
        col.child_at(1).append("y")
        col.set_length(2)

        self.assertEqual(col.type().name_of(0), "a")
        self.assertEqual(col.type().name_of(1), "b")
        self.assertEqual(col.child_at(col.type().get_child_idx("a"))[0], 1)
        self.assertEqual(col.child_at(col.type().get_child_idx("b"))[0], "x")
        self.assertEqual(col.child_at(col.type().get_child_idx("a"))[1], 2)
        self.assertEqual(col.child_at(col.type().get_child_idx("b"))[1], "y")

        sliced_col = col.slice(1, 1)
        self.assertEqual(
            sliced_col.child_at(sliced_col.type().get_child_idx("a"))[0], 2
        )
        self.assertEqual(
            sliced_col.child_at(sliced_col.type().get_child_idx("b"))[0], "y"
        )

    def test_set_child(self):
        col = ta.Column(
            ta.VeloxRowType(
                ["a", "b"], [ta.VeloxType_INTEGER(), ta.VeloxType_VARCHAR()]
            )
        )
        col.child_at(0).append(1)
        col.child_at(1).append("x")
        col.set_length(1)
        col.child_at(0).append(2)
        col.child_at(1).append("y")
        col.set_length(2)

        new_child = infer_column([3, 4])
        col.set_child(0, new_child)

        self.assertEqual(col.type().name_of(0), "a")
        self.assertEqual(col.type().name_of(1), "b")
        self.assertEqual(col.child_at(col.type().get_child_idx("a"))[0], 3)
        self.assertEqual(col.child_at(col.type().get_child_idx("b"))[0], "x")
        self.assertEqual(col.child_at(col.type().get_child_idx("a"))[1], 4)
        self.assertEqual(col.child_at(col.type().get_child_idx("b"))[1], "y")

    def test_nested_row(self):
        col = ta.Column(
            ta.VeloxRowType(
                ["a", "b"],
                [
                    ta.VeloxType_INTEGER(),
                    ta.VeloxRowType(
                        ["b1", "b2"], [ta.VeloxType_VARCHAR(), ta.VeloxType_INTEGER()]
                    ),
                ],
            )
        )
        col.child_at(0).append(1)
        col.child_at(1).child_at(0).append("21")
        col.child_at(1).child_at(1).append(22)
        self.assertEqual(col.type().get_child_idx("a"), 0)
        self.assertEqual(col.type().get_child_idx("b"), 1)
        self.assertEqual(col.child_at(1).type().get_child_idx("b1"), 0)
        self.assertEqual(col.child_at(1).type().get_child_idx("b2"), 1)
        self.assertEqual(col.child_at(0)[0], 1)
        self.assertEqual(col.child_at(1).child_at(0)[0], "21")
        self.assertEqual(col.child_at(1).child_at(1)[0], 22)


if __name__ == "__main__":
    unittest.main()
